/*
 * WMAL - Window Maker Application Launcher
 * Version:  0.1.2
 * Copyright (C) 1999 Martin Edlman
 *
 * Based on Casey Harkin's  WMAppl - Window Maker Application Launcher
 * Version:  0.2
 * Copyright (C) 1999 Casey Harkins
 * Contributors:
 *		Casey Harkins (charkins@pobox.com)
 *		Richard Eisenman 
 *		Jacob Beardsley	
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <Xlib.h>
#include <Xutil.h>
#include <extensions/shape.h>
#include <xpm.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include "interface.xpm"

#define APPNAME   "wmal"
#define CLASSNAME "WMApplicationLauncher"
#define VERSION   "0.1.2"

#ifndef DEFAULT_FONTNAME
#  define DEFAULT_FONTNAME "-adobe-helvetica-medium-r-normal-*-*-80-*-*-p-*-iso8859-1"
#endif

#ifndef DEFAULT_WMALPATH
#  define DEFAULT_WMALPATH   "/usr/local/wmal"
#endif

#define DEFAULT_ICONPATH   DEFAULT_WMALPATH"/icons"
#define DEFAULT_CONFIGFILE DEFAULT_WMALPATH"/wmalrc"

typedef struct _XpmButton {
	Pixmap pixmap;
        int pressed;
        char *label;
	char *command;
	struct _XpmButton *prev;
	struct _XpmButton *next;
} XpmButton;


/* some globals */
Display *display;
int screen;
Window rootwin;
Window win;
Window iconwin;
GC gc;
Pixmap pixmap, pixmask;
XpmAttributes pixattr;
XpmButton *icons, sup, sdown;
char *appname, *fontname, *classname;
int debugmode=0;
int button_positions[] = {
    4,  5, 51, 22,
    4, 23, 51, 40,
    4, 41, 51, 58
};

int scroll_positions[] = {
    52,  5, 59, 12,
    52, 51, 59, 58
};

/* x initialization crap */
void initx(int argc, char **argv) {
	int i;

	/* some x structs */
	XSizeHints xsh;
	XWMHints *xwmh;
	XClassHint xch;
	XGCValues xgcv;
        XTextProperty xtp;
        XFontStruct *font;
	
	/* connect to default display */
	display = XOpenDisplay(getenv("DISPLAY"));
	if(!display) {
		fprintf(stderr, "Couldn't connect to display\n");
		exit(1);
	}

	/* get screen and root window */
	screen = DefaultScreen(display);
	rootwin = RootWindow(display, screen);

	/* set size hints 64 x 64 */
	xsh.flags = USSize | USPosition;
	xsh.width = 64;
	xsh.height = 64;
	XWMGeometry(display, screen, "64x64+0+0", NULL, 0,
		&xsh, &xsh.x, &xsh.y, &xsh.width, &xsh.height, &i);

	/* create the application window */
	win = XCreateSimpleWindow(display, rootwin,
		xsh.x, xsh.y, xsh.width, xsh.height, 0,
		BlackPixel(display, screen),
		WhitePixel(display, screen));

	if(!win) {
		fprintf(stderr, "Couldn't create window\n");
		exit(1);
	}

	/* create icon window */
	iconwin = XCreateSimpleWindow(display, rootwin,
		xsh.x, xsh.y, xsh.width, xsh.height, 0,
		BlackPixel(display, screen),
		WhitePixel(display, screen));

	if(!iconwin) {
		fprintf(stderr, "Couldn't create icon window\n");
		exit(1);
	}


	/* load interface pixmap */
	if(XpmCreatePixmapFromData(display, rootwin, interface_xpm,
		&pixmap, &pixmask, &pixattr)!=XpmSuccess) {
		fprintf(stderr, "Couldn't create interface\n");
		exit(1);
	}

        sup.pressed=0;
        sdown.pressed=0;

	/* set background pixmap for iconwin */
	XSetWindowBackgroundPixmap(display, win, pixmap);
	XSetWindowBackgroundPixmap(display, iconwin, pixmap);

	/* should query server to see if Shape extension is supported */

	/* setup shaped window */
	XShapeCombineMask(display, win, ShapeBounding, 
		0, 0, pixmask, ShapeSet);
	XShapeCombineMask(display, iconwin, ShapeBounding,
		0, 0, pixmask, ShapeSet);

	/* set window manager hints */
	xwmh = XAllocWMHints();
	xwmh->flags = WindowGroupHint | IconWindowHint | StateHint;
	xwmh->icon_window = iconwin;
	xwmh->window_group = win;
	xwmh->initial_state = WithdrawnState;
	XSetWMHints(display, win, xwmh);

	
	/* set class hints */
	classname =(char *)malloc(strlen(CLASSNAME));
	strcpy(classname, CLASSNAME);
	xch.res_name = appname; 
	xch.res_class = classname;
	XSetClassHint(display, win, &xch);

	/* set size hints */
	XSetWMNormalHints(display, win, &xsh);
	
	/* tell window manager app name */
	if(!XStringListToTextProperty(&appname, 1, &xtp)) {
		fprintf(stderr, "Couldn't create text property\n");
		exit(1);
	}
	XSetWMName(display, win, &xtp);


	/* create a graphics context */
	gc = XCreateGC(display, win, GCForeground | GCBackground, &xgcv);
	if(!gc) {
		fprintf(stderr, "Couldn't create graphics context\n");
		exit(1);
        }
        font = XLoadQueryFont(display, fontname);
        if(font)
            XSetFont(display, gc, font->fid);

	/* select events to catch */
	XSelectInput(display, win,
		ExposureMask |
		ButtonPressMask |
		ButtonReleaseMask |
		PointerMotionMask |
		StructureNotifyMask);
	XSelectInput(display, iconwin,
		ExposureMask |
		ButtonPressMask |
		ButtonReleaseMask |
		PointerMotionMask |
		StructureNotifyMask);

	/* set the command line for restarting */
	XSetCommand(display, win, argv, argc);

	/* map the main window */
	XMapWindow(display, win);

}


void draw_actionbutton(XpmButton *b, int x, int y) {
    Pixel bl, wh;

    bl = BlackPixel(display, screen);
    wh = WhitePixel(display, screen);

    if(b->pressed)
        XSetForeground(display, gc, bl);
    else
        XSetForeground(display, gc, wh);

    XDrawLine(display, win, gc, x, y, x+47, y);
    XDrawLine(display, win, gc, x, y, x, y+17);

    XDrawLine(display, iconwin, gc, x, y, x+47, y);
    XDrawLine(display, iconwin, gc, x, y, x, y+17);

    if(b->pressed)
        XSetForeground(display, gc, wh);
    else
        XSetForeground(display, gc, bl);

    XDrawLine(display, win, gc, x+1, y+17, x+47, y+17);
    XDrawLine(display, win, gc, x+47, y+1, x+47, y+17);

    XDrawLine(display, iconwin, gc, x+1, y+17, x+47, y+17);
    XDrawLine(display, iconwin, gc, x+47, y+1, x+47, y+17);

    XCopyArea(display, b->pixmap, win, gc, 0, 0, 16, 16, x+1, y+1);
    XCopyArea(display, b->pixmap, iconwin, gc, 0, 0, 16, 16, x+1, y+1);

    if(debugmode)
        fprintf(stderr, "   label: %s (%i)\ncommand: %s (%i)\n",
                            b->label, strlen(b->label),
                            b->command, strlen(b->command));
    XSetForeground(display, gc, bl);
    XDrawString(display, win, gc, x+18, y+14, b->label, strlen(b->label));
    XDrawString(display, iconwin, gc, x+18, y+14, b->label, strlen(b->label));
}


void draw_scrollbutton(XpmButton *b, int x, int y) {
    Pixel bl, wh;

    bl = BlackPixel(display, screen);
    wh = WhitePixel(display, screen);

    if(b->pressed)
        XSetForeground(display, gc, bl);
    else
        XSetForeground(display, gc, wh);

    XDrawLine(display, win, gc, x, y, x+7, y);
    XDrawLine(display, win, gc, x, y, x, y+7);

    XDrawLine(display, iconwin, gc, x, y, x+7, y);
    XDrawLine(display, iconwin, gc, x, y, x, y+7);

    if(b->pressed)
        XSetForeground(display, gc, wh);
    else
        XSetForeground(display, gc, bl);

    XDrawLine(display, win, gc, x+1, y+7, x+7, y+7);
    XDrawLine(display, win, gc, x+7, y+1, x+7, y+7);

    XDrawLine(display, iconwin, gc, x+1, y+7, x+7, y+7);
    XDrawLine(display, iconwin, gc, x+7, y+1, x+7, y+7);
}

void draw(void) {
	XpmButton *p;
	int i;

	for(i=0, p=icons; i < 3 && p!=NULL; i++, p=p->next)
		draw_actionbutton(p, button_positions[i*4], button_positions[i*4+1]);

	draw_scrollbutton(&sup, scroll_positions[0], scroll_positions[1]);
	draw_scrollbutton(&sdown, scroll_positions[4], scroll_positions[5]);

}

XpmButton *which_button(int x, int y) {
    int i;
    XpmButton *p;

    for(i=0, p=icons; i<3 && p!=NULL; i++, p=p->next) {
        if(x>button_positions[i*4] && y>button_positions[i*4+1] &&
           x<button_positions[i*4+2] && y<button_positions[i*4+3])
            return p;
    }
    if(x>scroll_positions[0] && y>scroll_positions[1] &&
       x<scroll_positions[2] && y<scroll_positions[3])
        return &sup;
    if(x>scroll_positions[4] && y>scroll_positions[5] &&
       x<scroll_positions[6] && y<scroll_positions[7])
        return &sdown;
    return NULL;
}


void event_loop(void) {
    XEvent e;
    XpmButton *b = NULL;
    int i=1;

    while(i) {
        /* get next event from queue */
        XNextEvent(display, &e);
        /* handle events */
        switch(e.type) {

            /* Expose event */
        case Expose:
            if(e.xexpose.count!=0) break;
            draw();
            break;

            /* Configure Notify event */
        case ConfigureNotify:
            draw();
            break;

            /* Destroy Notify event */
        case DestroyNotify:
            i=0;
            break;

            /* Button Press event */
        case ButtonPress:
            if(e.xbutton.button != Button1) break;
            b = which_button(e.xbutton.x, e.xbutton.y);
            if(b != NULL) b->pressed = 1;
            draw();
            break;

            /* Button Release event */
        case ButtonRelease:
            if(e.xbutton.button != Button1) break;
            if(b != NULL && b == which_button(e.xbutton.x, e.xbutton.y)) {
                /* scroll up */
                if(b == &sup) {
                    if(icons->prev != NULL)
                        icons=icons->prev;
                    XClearWindow(display, win);
                    XClearWindow(display, iconwin);
                }
                /* scroll down - this is messy but nescessary*/
                else if(b == &sdown) {
                    /* the following nested if is not indented
                     * properly but at least it < 80 chars wide */
                    if(icons->next != NULL) {
                        if(icons->next->next != NULL) {
                            if(icons->next->next->next != NULL)
                                icons=icons->next;
                        }}
                    /* that's the end of the ugly nested if */
                    XClearWindow(display, win);
                    XClearWindow(display, iconwin);
                }
                /* execute command */
                else {
                    if(b->command!=NULL) {
                        if(debugmode)
                            fprintf(stderr,"%s\n",b->command);
                        else
                            system(b->command);
                    }
                }
            }
            if(b!=NULL) b->pressed = 0;
            draw();
            break;
        } // end event switch
    } // end while
} // end event_loop()


char *get_line(FILE *in) {
    char *line, *extra, *grow;
    int length=1024, i;

    line=(char *)malloc(sizeof(char)*length);
    if(!line) {
        fprintf(stderr, "Couldn't allocate buffer.\n");
        exit(1);
    }
    /* read in a line */
    if(!fgets(line, length, in)) {
        free(line);
        return NULL;
    }
    /* process the string */
    for(i=0; i<strlen(line); i++) {
        /* end string at a comment character */
        if(line[i]=='#') line[i]='\0';
        /* cat next line at an expansion character */
        if(line[i]=='\\') {
            line[i]='\0';
            extra=get_line(in);
            grow=(char *)malloc(sizeof(char)*(strlen(line)+strlen(extra)));
            strcpy(grow, line);
            strcat(grow, extra);
            free(line);
            free(extra);
            line=grow;
        }
        /* end string at a newline character */
        if(line[i]=='\n') line[i]='\0';
    }
    return line;
}



void loadIcons(char *filename, char *iconpath) {
    FILE *f;
    XpmButton *l, *c;
    char *line, *icon, *label, *command, *buf;
    int first = 1, i = 0, ignore = 0;

    f = fopen(filename, "r");
    if(f == NULL) {
        fprintf(stderr, "Couldn't open file %s for reading.\n", filename);
        exit(1);
    }

    l = NULL;

    /* read a line */
    line = get_line(f);

    while(line) {
        if(strlen(line) > 0) { // non-empty line

            i = ignore = 0;
            /* get the icon filename */
            icon = line;
            while(line[i] != '\0' && line[i] != ' ' && line[i] != '\t') i++;
            if(line[i] == '\0') {
                fprintf(stderr, "Error parsing line (iconname): %s\n", line);
                ignore = 1;
            }
            line[i++] = '\0';  // terminate icon part
            /* skip blanks */
            while(line[i] != '\0' && (line[i] == ' ' || line[i] == '\t')) i++;
            if(line[i] == '\0') {
                fprintf(stderr, "Error parsing line (blanks after iconname): %s\nIconname: %s", line, icon);
                ignore = 1;
            }
            /* read label */
            label = &line[i];
            while(line[i] != '\0' && line[i] != ' ' && line[i] != '\t') i++;
            if(line[i] == '\0') {
                fprintf(stderr, "Error parsing line (label): %s\nIcon: %s", line, icon);
                ignore = 1;
            }
            line[i++] = '\0';  // terminate label part
            /* skip blanks */
            while(line[i] != '\0' && (line[i] == ' ' || line[i] == '\t')) i++;
            if(line[i] == '\0') {
                fprintf(stderr, "Error parsing line (blanks after label): %s\n", line);
                ignore = 1;
            }
            /* command is the rest of line */
            command = &line[i];

            if(ignore)
                fprintf(stderr, "Ignoring line...\n");
            else {
                c = (XpmButton*)malloc(sizeof(XpmButton));
                if(first > 0) {
                    icons = c;
                    first = 0;
                }
                c->pressed = 0;
                c->prev = l;
                if(l != NULL) l->next = c;
                c->next = NULL;
        
                /* prepend icon path */
                buf = (char *)malloc(sizeof(char)*(strlen(icon)+strlen(iconpath)+2));
                if(!buf) {
                    fprintf(stderr, "Couldn't allocate string.\n");
                    exit(1);
                }
                strcpy(buf, iconpath);
                i=strlen(buf);
                if(i > 0 && buf[i-1] != '/')
                    strcat(buf,"/");
                strcat(buf, icon);
                strcat(buf,"\0");
                if(debugmode) {
                    fprintf(stderr, "Loading icon %s.\n",buf);
                }
                if(XpmReadFileToPixmap(display, win, buf, &c->pixmap, NULL, NULL)
                   !=XpmSuccess) {
                    fprintf(stderr, "Error loading icon %s.\n", buf);
                    exit(1);
                }
                c->label = (char*)malloc(strlen(label)*sizeof(char) + 1);
                strcpy(c->label, label);
                strcat(c->label, "\0");
                c->command = (char*)malloc(strlen(command)*sizeof(char) + 1);
                strcpy(c->command, command);
                strcat(c->label, "\0");
                free(line);
                if(debugmode) {
                    fprintf(stderr, "   label: %s (%i)\nc->label: %s (%i)\n",
                            label, strlen(label), c->label, strlen(c->label));
                    fprintf(stderr, "   commands: %s (%i)\nc->command: %s (%i)\n",
                            command, strlen(command), c->command, strlen(c->command));
                }
        
                l=c;
            }                 // end if(ignore)
        }                     // end if(strlen(line) > 0)
        line=get_line(f);
    }
    fclose(f);
}

void show_usage() {
    fprintf(stderr, "\nUsage:\
wmal [--config configfile] [--name appname] [--font fontname] [--icons iconpath] [--debug] [--verbose]\n\
  --config  -c: reads configfile rather than ~/.wmalrc\n\
  --name    -n: set application name, useful for running mutliple instances\n\
  --font    -f: use specified font for labels in XLFD format\n\t    default: %s\n\
  --icons   -i: prefixes icon filenames with iconpath\n\t    default: %s\n\
  --debug   -d: debug mode (doesn't execute commands)\n\
  --verbose -V: show version information\n",
    DEFAULT_FONTNAME, DEFAULT_ICONPATH);
    exit(1);
}

void show_version() {
    fprintf(stderr, "\nwmal v%s\n", VERSION);
    fprintf(stderr, "(http://edas.visaci.cz/src/#wmal)\n");
    exit(1);
}

int main(int argc, char **argv) {
    int  option_index, c;
    char configfile[1024];	// config file to read from
    char iconpath[1024];	// path to load icons from
    FILE *f;

    static struct option long_options[] = {
        {"config", required_argument, 0, 'c'},
        {"name", required_argument, 0, 'n'},
        {"font", required_argument, 0, 'f'},
        {"icons", required_argument, 0, 'i'},
        {"debug", no_argument, 0, 'd'},
        {"verbose", no_argument, 0, 'V'},
        {0, 0, 0, 0}
    };
                                                                                                                                                    
    /* set default app class name */
    appname = (char*)malloc(sizeof(char)*1024);
    strncpy(appname, APPNAME, 1023);

    /* set default config file */
    strcpy(configfile, getenv("HOME"));
    strcat(configfile, "/.wmalrc");

    /* set default font name */
    fontname = (char*)malloc(sizeof(char)*1024);
    strncpy(fontname, DEFAULT_FONTNAME, 1023);

    /* set default icon path */
    strncpy(iconpath, DEFAULT_ICONPATH, 1023);

    /* parse command line arguments */
    while(1) {
        c = getopt_long(argc, argv, "c:n:f:i:dV", long_options, &option_index);
        if(c == -1) 
            break;
        switch(c) {
            /* specify config file */
            case 'c':
                strncpy(configfile, optarg, 1023);
                break;

            /* specify application name */
            case 'n':
                strncpy(appname, optarg, 1023);
                break;

            /* specify font name path */
            case 'f':
                strncpy(fontname, optarg, 1023);
                break;

            /* specify icon path */
            case 'i':
                strncpy(iconpath, optarg, 1023);
                break;

            /* set debug mode */
            case 'd':
                debugmode=1;
                break;

            /* show version information */
            case 'V':
                show_version();
                break;

            /* anything else show usage information */
            default:
                show_usage();
                break;

        } // end switch
    } // end while

    f = fopen(configfile, "r");
    if(f == NULL) {
        fprintf(stderr, "Couldn't open file %s for reading.\nResetting to %s\n", configfile, DEFAULT_CONFIGFILE);
        strcpy(configfile, DEFAULT_CONFIGFILE);
    }
    else
        fclose(f);

    if(debugmode)
        fprintf(stderr, "FN: %s\nIP: %s\nCF: %s\n\n", fontname, iconpath, configfile);

    initx(argc, argv);
    loadIcons(configfile, iconpath);
    event_loop();
    free(fontname);
    return 0;
}
